@@ -20,10 +20,13 @@
 #include "qgis_sip.h"
 #include "qgsattributeeditorcontext.h"
 #include "qgseditorwidgetwrapper.h"
+#include "qgsattributeeditorelement.h"
 
 #include <QWidget>
 #include <QLabel>
 #include <QDialogButtonBox>
+#include <QMultiMap>
+
 #include "qgis_gui.h"
 
 
@@ -55,8 +58,8 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
       will add a new feature when the form is accepted. */
       MultiEditMode, //!< Multi edit mode, for editing fields of multiple features at once
       SearchMode, //!< Form values are used for searching/filtering the layer
-      AggregateSearchMode, //!< Form is in aggregate search mode, show each widget in this mode \since QGIS 3.0
-      IdentifyMode //!< Identify the feature \since QGIS 3.0
+      AggregateSearchMode, //!< Form is in aggregate search mode, show each widget in this mode
+      IdentifyMode //!< Identify the feature
     };
 
     //! Filter types
@@ -138,7 +141,6 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
     /**
      * Returns the current mode of the form.
      * \see setMode()
-     * \since QGIS 2.16
      */
     QgsAttributeEditorContext::Mode mode() const { return mMode; }
 
@@ -146,7 +148,6 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
      * Sets the current mode of the form.
      * \param mode form mode
      * \see mode()
-     * \since QGIS 2.16
      */
     void setMode( QgsAttributeEditorContext::Mode mode );
 
@@ -170,7 +171,6 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
     /**
      * Sets all feature IDs which are to be edited if the form is in multiedit mode
      * \param fids feature ID list
-     * \since QGIS 2.16
      */
     void setMultiEditFeatureIds( const QgsFeatureIds &fids );
 
@@ -178,7 +178,6 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
      * Sets the message bar to display feedback from the form in. This is used in the search/filter
      * mode to display the count of selected features.
      * \param messageBar target message bar
-     * \since QGIS 2.16
      */
     void setMessageBar( QgsMessageBar *messageBar );
 
@@ -187,7 +186,6 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
      * In this case it will return a combined expression according to the chosen filters
      * on all attribute widgets.
      *
-     * \since QGIS 3.0
      */
     QString aggregateFilter() const;
 
@@ -199,6 +197,12 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
      */
     void setExtraContextScope( QgsExpressionContextScope *extraScope SIP_TRANSFER );
 
+    /**
+     * Returns TRUE if any of the form widgets need feature geometry
+     * \since QGIS 3.20
+     */
+    bool needsGeometry() const;
+
   signals:
 
     /**
@@ -217,7 +221,6 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
      * \param attribute The name of the attribute that changed.
      * \param value     The new value of the attribute.
      * \param attributeChanged If TRUE, it corresponds to an actual change of the feature attribute
-     * \since QGIS 3.0.1
      */
     void widgetValueChanged( const QString &attribute, const QVariant &value, bool attributeChanged );
 
@@ -240,7 +243,6 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
      * Emitted when a filter expression is set using the form.
      * \param expression filter expression
      * \param type filter type
-     * \since QGIS 2.16
      */
     void filterExpressionSet( const QString &expression, QgsAttributeForm::FilterType type );
 
@@ -252,22 +254,25 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
 
     /**
      * Emitted when the user selects the close option from the form's button bar.
-     * \since QGIS 2.16
      */
     void closed();
 
     /**
      * Emitted when the user chooses to zoom to a filtered set of features.
-     * \since QGIS 3.0
      */
     void zoomToFeatures( const QString &filter );
 
     /**
      * Emitted when the user chooses to flash a filtered set of features.
-     * \since QGIS 3.0
      */
     void flashFeatures( const QString &filter );
 
+    /**
+     * Emitted when the user chooses to open the attribute table dialog with a filtered set of features.
+     * \since QGIS 3.24
+     */
+    void openFilteredFeaturesAttributeTable( const QString &filter );
+
   public slots:
 
     /**
@@ -279,6 +284,13 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
      */
     void changeAttribute( const QString &field, const QVariant &value, const QString &hintText = QString() );
 
+    /**
+     * Changes the \a geometry of the feature attached to the form.
+     *
+     * \since QGIS 3.30
+     */
+    void changeGeometry( const QgsGeometry &geometry );
+
     /**
      * Update all editors to correspond to a different feature.
      *
@@ -311,7 +323,6 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
 
     /**
      * Resets the search/filter form values.
-     * \since QGIS 2.16
      */
     void resetSearch();
 
@@ -331,16 +342,11 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
      */
     void parentFormValueChanged( const QString &attribute, const QVariant &newValue );
 
-    /**
-     * Returns TRUE if any of the form widgets need feature geometry
-     * \since QGIS 3.20
-     */
-    bool needsGeometry() const;
-
   private slots:
     void onAttributeChanged( const QVariant &value, const QVariantList &additionalFieldValues );
     void onAttributeAdded( int idx );
     void onAttributeDeleted( int idx );
+    void onRelatedFeaturesChanged();
     void onUpdatedFields();
     void onConstraintStatusChanged( const QString &constraint,
                                     const QString &description, const QString &err, QgsEditorWidgetWrapper::ConstraintResult result );
@@ -380,6 +386,9 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
     void updateFieldDependencies();
     void updateFieldDependenciesDefaultValue( QgsEditorWidgetWrapper *eww );
     void updateFieldDependenciesVirtualFields( QgsEditorWidgetWrapper *eww );
+    void updateRelatedLayerFieldsDependencies( QgsEditorWidgetWrapper *eww = nullptr );
+
+    void setMultiEditFeatureIdsRelations( const QgsFeatureIds &fids );
 
     struct WidgetInfo
     {
@@ -390,12 +399,11 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
       bool labelOnTop = false;
       bool labelAlignRight = false;
       bool showLabel = true;
+      QgsAttributeEditorElement::LabelStyle labelStyle;
     };
 
     WidgetInfo createWidgetFromDef( const QgsAttributeEditorElement *widgetDef, QWidget *parent, QgsVectorLayer *vl, QgsAttributeEditorContext &context );
 
-    void addWidgetWrapper( QgsEditorWidgetWrapper *eww );
-
     /**
      * Creates widget wrappers for all suitable widgets found.
      * Called once maximally.
@@ -408,10 +416,13 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
     //! Save single feature or add feature edits
     bool saveEdits( QString *error );
 
+    QgsFeature getUpdatedFeature() const;
+
     //! update the default values and virtual fields in the fields after a referenced field changed
     void updateValuesDependencies( const int originIdx );
     void updateValuesDependenciesDefaultValues( const int originIdx );
     void updateValuesDependenciesVirtualFields( const int originIdx );
+    void updateRelatedLayerFields();
 
     void clearMultiEditMessages();
     void pushSelectedFeaturesMessage();
@@ -427,8 +438,10 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
     void updateContainersVisibility();
     void updateConstraint( const QgsFeature &ft, QgsEditorWidgetWrapper *eww );
     void updateLabels();
+    void updateEditableState();
     bool currentFormValuesFeature( QgsFeature &feature );
-    bool currentFormValidConstraints( QStringList &invalidFields, QStringList &descriptions );
+    bool currentFormValidConstraints( QStringList &invalidFields, QStringList &descriptions ) const;
+    bool currentFormValidHardConstraints( QStringList &invalidFields, QStringList &descriptions ) const;
     QList<QgsEditorWidgetWrapper *> constraintDependencies( QgsEditorWidgetWrapper *w );
 
     Q_DECL_DEPRECATED QgsRelationWidgetWrapper *setupRelationWidgetWrapper( const QgsRelation &rel, const QgsAttributeEditorContext &context ) SIP_DEPRECATED;
@@ -447,10 +460,11 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
     QDialogButtonBox *mButtonBox = nullptr;
     QWidget *mSearchButtonBox = nullptr;
     QList<QgsAttributeFormInterface *> mInterfaces;
-    QMap< int, QgsAttributeFormEditorWidget * > mFormEditorWidgets;
+    QMultiMap< int, QgsAttributeFormEditorWidget * > mFormEditorWidgets;
     QList< QgsAttributeFormWidget *> mFormWidgets;
     QMap<const QgsVectorLayerJoinInfo *, QgsFeature> mJoinedFeatures;
     QMap<QLabel *, QgsProperty> mLabelDataDefinedProperties;
+    QMap<QWidget *, QgsProperty> mEditableDataDefinedProperties;
     bool mValuesInitialized = false;
     bool mDirty = false;
     bool mIsSettingFeature = false;
@@ -473,10 +487,21 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
         , isVisible( true )
       {}
 
+      ContainerInformation( QWidget *widget, const QgsExpression &visibilityExpression, bool collapsed, const QgsExpression &collapsedExpression )
+        : widget( widget )
+        , expression( visibilityExpression )
+        , isVisible( true )
+        , isCollapsed( collapsed )
+        , collapsedExpression( collapsedExpression )
+      {}
+
+
       QgsTabWidget *tabWidget = nullptr;
       QWidget *widget = nullptr;
       QgsExpression expression;
       bool isVisible;
+      bool isCollapsed = false;
+      QgsExpression collapsedExpression;
 
       void apply( QgsExpressionContext *expressionContext );
     };
@@ -487,8 +512,8 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
 
     void reloadIcon( const QString &file, const QString &tooltip, QSvgWidget *sw );
 
-    // Contains information about tabs and groupboxes, their visibility state visibility conditions
-    QVector<ContainerInformation *> mContainerVisibilityInformation;
+    // Contains information about tabs and groupboxes, their visibility/collapsed state conditions
+    QVector<ContainerInformation *> mContainerVisibilityCollapsedInformation;
     QMap<QString, QVector<ContainerInformation *> > mContainerInformationDependency;
 
     // Variables below are used for Python
@@ -517,13 +542,18 @@ class GUI_EXPORT QgsAttributeForm : public QWidget
      * Dependency map for default values. Attribute index -> widget wrapper.
      * Attribute indexes will be added multiple times if more than one widget depends on them.
      */
-    QMap<int, QgsWidgetWrapper *> mDefaultValueDependencies;
+    QMultiMap<int, QgsWidgetWrapper *> mDefaultValueDependencies;
 
     /**
      * Dependency map for values from virtual fields. Attribute index -> widget wrapper.
      * Attribute indexes will be added multiple times if more than one widget depends on them.
      */
-    QMap<int, QgsWidgetWrapper *> mVirtualFieldsDependencies;
+    QMultiMap<int, QgsWidgetWrapper *> mVirtualFieldsDependencies;
+
+    /**
+     * Dependency list for values depending on related layers.
+     */
+    QSet<QgsEditorWidgetWrapper *> mRelatedLayerFieldsDependencies;
 
     //! List of updated fields to avoid recursion on the setting of defaultValues
     QList<int> mAlreadyUpdatedFields;
