@@ -20,10 +20,12 @@
 #include "qgsrendererwidget.h"
 #include "qgsproxystyle.h"
 #include <QStandardItem>
+#include <QStyledItemDelegate>
 
 
 class QgsCategorizedSymbolRenderer;
 class QgsRendererCategory;
+class QgsSymbolSelectorWidget;
 
 #include "ui_qgscategorizedsymbolrendererwidget.h"
 #include "qgis_gui.h"
@@ -36,7 +38,7 @@ class GUI_EXPORT QgsCategorizedSymbolRendererModel : public QAbstractItemModel
 {
     Q_OBJECT
   public:
-    QgsCategorizedSymbolRendererModel( QObject *parent = nullptr );
+    QgsCategorizedSymbolRendererModel( QObject *parent = nullptr, QScreen *screen = nullptr );
     Qt::ItemFlags flags( const QModelIndex &index ) const override;
     Qt::DropActions supportedDropActions() const override;
     QVariant data( const QModelIndex &index, int role ) const override;
@@ -65,6 +67,7 @@ class GUI_EXPORT QgsCategorizedSymbolRendererModel : public QAbstractItemModel
   private:
     QgsCategorizedSymbolRenderer *mRenderer = nullptr;
     QString mMimeFormat;
+    QPointer< QScreen > mScreen;
 };
 
 /**
@@ -81,6 +84,26 @@ class QgsCategorizedSymbolRendererViewStyle: public QgsProxyStyle
     void drawPrimitive( PrimitiveElement element, const QStyleOption *option, QPainter *painter, const QWidget *widget = nullptr ) const override;
 };
 
+/**
+ * \ingroup gui
+ * \brief Custom delegate for localized numeric input.
+ */
+class QgsCategorizedRendererViewItemDelegate: public QStyledItemDelegate
+{
+    Q_OBJECT
+
+  public:
+    explicit QgsCategorizedRendererViewItemDelegate( QgsFieldExpressionWidget *expressionWidget, QObject *parent = nullptr );
+
+    // QAbstractItemDelegate interface
+    QWidget *createEditor( QWidget *parent, const QStyleOptionViewItem &option, const QModelIndex &index ) const override;
+
+  private:
+
+    QgsFieldExpressionWidget *mFieldExpressionWidget = nullptr;
+};
+
+
 ///@endcond
 
 #endif
@@ -93,6 +116,22 @@ class GUI_EXPORT QgsCategorizedSymbolRendererWidget : public QgsRendererWidget,
 {
     Q_OBJECT
   public:
+
+    // *INDENT-OFF*
+
+    /**
+     * Custom model roles.
+     *
+     * \note Prior to QGIS 3.36 this was available as QgsCategorizedSymbolRendererWidget::CustomRoles
+     * \since QGIS 3.36
+     */
+    enum class CustomRole SIP_MONKEYPATCH_SCOPEENUM_UNNEST( QgsCategorizedSymbolRendererWidget, CustomRoles ) : int
+    {
+      Value SIP_MONKEYPATCH_COMPAT_NAME(ValueRole) = Qt::UserRole + 1 //!< Category value
+    };
+    Q_ENUM( CustomRole )
+    // *INDENT-ON*
+
     static QgsRendererWidget *create( QgsVectorLayer *layer, QgsStyle *style, QgsFeatureRenderer *renderer ) SIP_FACTORY;
 
     QgsCategorizedSymbolRendererWidget( QgsVectorLayer *layer, QgsStyle *style, QgsFeatureRenderer *renderer );
@@ -100,6 +139,7 @@ class GUI_EXPORT QgsCategorizedSymbolRendererWidget : public QgsRendererWidget,
 
     QgsFeatureRenderer *renderer() override;
     void setContext( const QgsSymbolWidgetContext &context ) override;
+    void disableSymbolLevels() override SIP_SKIP;
 
     /**
      * Replaces category symbols with the symbols from a style that have a matching
@@ -108,7 +148,6 @@ class GUI_EXPORT QgsCategorizedSymbolRendererWidget : public QgsRendererWidget,
      * \returns number of symbols matched
      * \see matchToSymbolsFromLibrary
      * \see matchToSymbolsFromXml
-     * \since QGIS 2.9
      */
     int matchToSymbols( QgsStyle *style );
 
@@ -136,7 +175,6 @@ class GUI_EXPORT QgsCategorizedSymbolRendererWidget : public QgsRendererWidget,
      * matching name.
      * \see matchToSymbolsFromXml
      * \see matchToSymbols
-     * \since QGIS 2.9
      */
     void matchToSymbolsFromLibrary();
 
@@ -145,18 +183,19 @@ class GUI_EXPORT QgsCategorizedSymbolRendererWidget : public QgsRendererWidget,
      * from the XML file with a matching name.
      * \see matchToSymbolsFromLibrary
      * \see matchToSymbols
-     * \since QGIS 2.9
      */
     void matchToSymbolsFromXml();
 
+  protected:
+    void setSymbolLevels( const QgsLegendSymbolList &levels, bool enabled ) override;
+
   protected slots:
 
     void pasteSymbolToSelection() override;
 
   private slots:
 
-    void cleanUpSymbolSelector( QgsPanelWidget *container );
-    void updateSymbolsFromWidget();
+    void updateSymbolsFromWidget( QgsSymbolSelectorWidget *widget );
     void updateSymbolsFromButton();
     void dataDefinedSizeLegend();
 
@@ -216,6 +255,7 @@ class GUI_EXPORT QgsCategorizedSymbolRendererWidget : public QgsRendererWidget,
     QMenu *mContextMenu = nullptr;
     QAction *mMergeCategoriesAction = nullptr;
     QAction *mUnmergeCategoriesAction = nullptr;
+    QAction *mActionLevels = nullptr;
 
     QgsExpressionContext createExpressionContext() const override;
 
