@@ -22,11 +22,13 @@
 #include "qgis_sip.h"
 #include <QColor>
 #include <QPainter>
+#include <QPainterPath>
 #include <memory>
 
 #include "qgscoordinatetransform.h"
 #include "qgsexpressioncontext.h"
 #include "qgsfeaturefilterprovider.h"
+#include "qgslabelsink.h"
 #include "qgsmaptopixel.h"
 #include "qgsmapunitscale.h"
 #include "qgsrectangle.h"
@@ -34,7 +36,6 @@
 #include "qgsdistancearea.h"
 #include "qgscoordinatetransformcontext.h"
 #include "qgspathresolver.h"
-#include "qgssymbollayerreference.h"
 #include "qgstemporalrangeobject.h"
 
 class QPainter;
@@ -45,11 +46,13 @@ class QgsRenderedFeatureHandlerInterface;
 class QgsSymbolLayer;
 class QgsMaskIdProvider;
 class QgsMapClippingRegion;
+class QgsElevationMap;
 
 
 /**
  * \ingroup core
- * Contains information about the context of a rendering operation.
+ * \brief Contains information about the context of a rendering operation.
+ *
  * The context of a rendering operation defines properties such as
  * the conversion ratio between screen and map units, the extents
  * to be rendered etc.
@@ -63,102 +66,28 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
     QgsRenderContext( const QgsRenderContext &rh );
     QgsRenderContext &operator=( const QgsRenderContext &rh );
 
-    /**
-     * Enumeration of flags that affect rendering operations.
-     * \since QGIS 2.14
-     */
-    enum Flag
-    {
-      DrawEditingInfo          = 0x01,  //!< Enable drawing of vertex markers for layers in editing mode
-      ForceVectorOutput        = 0x02,  //!< Vector graphics should not be cached and drawn as raster images
-      UseAdvancedEffects       = 0x04,  //!< Enable layer opacity and blending effects
-      UseRenderingOptimization = 0x08,  //!< Enable vector simplification and other rendering optimizations
-      DrawSelection            = 0x10,  //!< Whether vector selections should be shown in the rendered map
-      DrawSymbolBounds         = 0x20,  //!< Draw bounds of symbols (for debugging/testing)
-      RenderMapTile            = 0x40,  //!< Draw map such that there are no problems between adjacent tiles
-      Antialiasing             = 0x80,  //!< Use antialiasing while drawing
-      RenderPartialOutput      = 0x100, //!< Whether to make extra effort to update map image with partially rendered layers (better for interactive map canvas). Added in QGIS 3.0
-      RenderPreviewJob         = 0x200, //!< Render is a 'canvas preview' render, and shortcuts should be taken to ensure fast rendering
-      RenderBlocking           = 0x400, //!< Render and load remote sources in the same thread to ensure rendering remote sources (svg and images). WARNING: this flag must NEVER be used from GUI based applications (like the main QGIS application) or crashes will result. Only for use in external scripts or QGIS server.
-      RenderSymbolPreview      = 0x800, //!< The render is for a symbol preview only and map based properties may not be available, so care should be taken to handle map unit based sizes in an appropriate way.
-      LosslessImageRendering   = 0x1000, //!< Render images losslessly whenever possible, instead of the default lossy jpeg rendering used for some destination devices (e.g. PDF). This flag only works with builds based on Qt 5.13 or later.
-      ApplyScalingWorkaroundForTextRendering = 0x2000, //!< Whether a scaling workaround designed to stablise the rendering of small font sizes (or for painters scaled out by a large amount) when rendering text. Generally this is recommended, but it may incur some performance cost.
-      Render3DMap              = 0x4000, //!< Render is for a 3D map
-      ApplyClipAfterReprojection = 0x8000, //!< Feature geometry clipping to mapExtent() must be performed after the geometries are transformed using coordinateTransform(). Usually feature geometry clipping occurs using the extent() in the layer's CRS prior to geometry transformation, but in some cases when extent() could not be accurately calculated it is necessary to clip geometries to mapExtent() AFTER transforming them using coordinateTransform().
-    };
-    Q_DECLARE_FLAGS( Flags, Flag )
-
-    /**
-     * Options for rendering text.
-     * \since QGIS 3.4.3
-     */
-    enum TextRenderFormat
-    {
-      // refs for below dox: https://github.com/qgis/QGIS/pull/1286#issuecomment-39806854
-      // https://github.com/qgis/QGIS/pull/8573#issuecomment-445585826
-
-      /**
-       * Always render text using path objects (AKA outlines/curves).
-       *
-       * This setting guarantees the best quality rendering, even when using a raster paint surface
-       * (where sub-pixel path based text rendering is superior to sub-pixel text-based rendering).
-       * The downside is that text is converted to paths only, so users cannot open created vector
-       * outputs for post-processing in other applications and retain text editability.
-       *
-       * This setting also guarantees complete compatibility with the full range of formatting options available
-       * through QgsTextRenderer and QgsTextFormat, some of which may not be possible to reproduce when using
-       * a vector-based paint surface and TextFormatAlwaysText mode.
-       *
-       * A final benefit to this setting is that vector exports created using text as outlines do
-       * not require all users to have the original fonts installed in order to display the
-       * text in its original style.
-       */
-      TextFormatAlwaysOutlines,
-
-      /**
-       * Always render text as text objects.
-       *
-       * While this mode preserves text objects as text for post-processing in external vector editing applications,
-       * it can result in rendering artifacts or poor quality rendering, depending on the text format settings.
-       *
-       * Even with raster based paint devices, TextFormatAlwaysText can result in inferior rendering quality
-       * to TextFormatAlwaysOutlines.
-       *
-       * When rendering using TextFormatAlwaysText to a vector based device (e.g. PDF or SVG), care must be
-       * taken to ensure that the required fonts are available to users when opening the created files,
-       * or default fallback fonts will be used to display the output instead. (Although PDF exports MAY
-       * automatically embed some fonts when possible, depending on the user's platform).
-       */
-      TextFormatAlwaysText,
-    };
-
     /**
      * Set combination of flags that will be used for rendering.
-     * \since QGIS 2.14
      */
-    void setFlags( QgsRenderContext::Flags flags );
+    void setFlags( Qgis::RenderContextFlags flags );
 
     /**
      * Enable or disable a particular flag (other flags are not affected)
-     * \since QGIS 2.14
      */
-    void setFlag( Flag flag, bool on = true );
+    void setFlag( Qgis::RenderContextFlag flag, bool on = true );
 
     /**
      * Returns combination of flags used for rendering.
-     * \since QGIS 2.14
      */
-    Flags flags() const;
+    Qgis::RenderContextFlags flags() const;
 
     /**
      * Check whether a particular flag is enabled.
-     * \since QGIS 2.14
      */
-    bool testFlag( Flag flag ) const;
+    bool testFlag( Qgis::RenderContextFlag flag ) const;
 
     /**
      * create initialized QgsRenderContext instance from given QgsMapSettings
-     * \since QGIS 2.4
      */
     static QgsRenderContext fromMapSettings( const QgsMapSettings &mapSettings );
 
@@ -166,7 +95,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * Creates a default render context given a pixel based QPainter destination.
      * If no painter is specified or the painter has no device, then a default
      * DPI of 88 will be assumed.
-     * \since QGIS 3.0
      */
     static QgsRenderContext fromQPainter( QPainter *painter );
 
@@ -188,6 +116,29 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
     const QPainter *painter() const { return mPainter; }
 #endif
 
+    /**
+     * Returns the const destination QPainter for temporary in-progress preview renders.
+     *
+     * May be NULLPTR if temporary in-progress preview renders are not required.
+     *
+     * \see setPreviewRenderPainter()
+     * \since QGIS 3.34
+    */
+    QPainter *previewRenderPainter() {return mPreviewRenderPainter;}
+
+#ifndef SIP_RUN
+
+    /**
+     * Returns the const destination QPainter for temporary in-progress preview renders.
+     *
+     * May be NULLPTR if temporary in-progress preview renders are not required.
+     *
+     * \see setPreviewRenderPainter()
+     * \since QGIS 3.34
+    */
+    const QPainter *previewRenderPainter() const { return mPreviewRenderPainter; }
+#endif
+
     /**
      * Sets relevant flags on a destination \a painter, using the flags and settings
      * currently defined for the render context.
@@ -210,6 +161,31 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      */
     QPainter *maskPainter( int id = 0 ) { return mMaskPainter.value( id, nullptr ); }
 
+    // TODO QGIS 4 : remove the V2 from method name
+
+    /**
+     * When rendering a map layer in a second pass (for selective masking),
+     * some symbol layers may be disabled.
+     *
+     * Sets the list of disabled symbol layers.
+     * \see disabledSymbolLayers()
+     * \see isSymbolLayerEnabled()
+     * \since QGIS 3.12
+     * \deprecated since QGIS 3.30 and replaced with setDisabledSymbolLayersV2
+     */
+    Q_DECL_DEPRECATED void setDisabledSymbolLayers( const QSet<const QgsSymbolLayer *> &symbolLayers ) SIP_DEPRECATED;
+
+    /**
+     * When rendering a map layer in a second pass (for selective masking),
+     * some symbol layers may be disabled.
+     *
+     * Sets the list of disabled symbol layer ids.
+     * \see disabledSymbolLayersV2()
+     * \see isSymbolLayerEnabled()
+     * \since QGIS 3.30
+     */
+    void setDisabledSymbolLayersV2( const QSet<QString> &symbolLayers );
+
     /**
      * When rendering a map layer in a second pass (for selective masking),
      * some symbol layers may be disabled.
@@ -218,8 +194,20 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * \see setDisabledSymbolLayers()
      * \see isSymbolLayerEnabled()
      * \since QGIS 3.12
+     * \deprecated since QGIS 3.30 and replaced with disabledSymbolLayersV2
      */
-    QSet<const QgsSymbolLayer *> disabledSymbolLayers() const { return mDisabledSymbolLayers; }
+    Q_DECL_DEPRECATED QSet<const QgsSymbolLayer *> disabledSymbolLayers() const SIP_DEPRECATED;
+
+    /**
+     * When rendering a map layer in a second pass (for selective masking),
+     * some symbol layers may be disabled.
+     *
+     * Returns the list of disabled symbol layer ids.
+     * \see setDisabledSymbolLayers()
+     * \see isSymbolLayerEnabled()
+     * \since QGIS 3.30
+     */
+    QSet<QString> disabledSymbolLayersV2() const;
 
     /**
      * When rendering a map layer in a second pass (for selective masking),
@@ -230,7 +218,7 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * \see disabledSymbolLayers()
      * \since QGIS 3.12
      */
-    bool isSymbolLayerEnabled( const QgsSymbolLayer *layer ) const { return ! mDisabledSymbolLayers.contains( layer ); }
+    bool isSymbolLayerEnabled( const QgsSymbolLayer *layer ) const;
 
     /**
      * Returns the current coordinate transform for the context.
@@ -246,7 +234,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
 
     /**
      * A general purpose distance and area calculator, capable of performing ellipsoid based calculations.
-     * \since QGIS 3.0
      */
     const QgsDistanceArea &distanceArea() const { return mDistanceArea; }
 
@@ -256,7 +243,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * from a source to destination coordinate reference system.
      *
      * \see setTransformContext()
-     * \since QGIS 3.0
      */
     QgsCoordinateTransformContext transformContext() const;
 
@@ -266,7 +252,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * from a source to destination coordinate reference system.
      *
      * \see transformContext()
-     * \since QGIS 3.0
      */
     void setTransformContext( const QgsCoordinateTransformContext &context );
 
@@ -275,7 +260,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * during rendering operations, e.g. for resolving relative symbol paths.
      *
      * \see setPathResolver()
-     * \since QGIS 3.0
      */
     const QgsPathResolver &pathResolver() const { return mPathResolver; }
 
@@ -284,7 +268,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * during rendering operations, e.g. for resolving relative symbol paths.
      *
      * \see pathResolver()
-     * \since QGIS 3.0
      */
     void setPathResolver( const QgsPathResolver &resolver ) { mPathResolver = resolver; }
 
@@ -332,14 +315,48 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      */
     double scaleFactor() const {return mScaleFactor;}
 
+    /**
+     * Returns the targeted DPI for rendering.
+     *
+     * \see setDpiTarget()
+     * \since QGIS 3.20
+     */
+    double dpiTarget() const {return mDpiTarget;}
+
     /**
      * Returns TRUE if the rendering operation has been stopped and any ongoing
      * rendering should be canceled immediately.
      *
+     * \note Since QGIS 3.22 the feedback() member exists as an alternative means of cancellation support.
+     *
      * \see setRenderingStopped()
+     * \see feedback()
      */
     bool renderingStopped() const {return mRenderingStopped;}
 
+    /**
+     * Attach a \a feedback object that can be queried regularly during rendering to check
+     * if rendering should be canceled.
+     *
+     * Ownership of \a feedback is NOT transferred, and the caller must take care that it exists
+     * for the lifetime of the render context.
+     *
+     * \see feedback()
+     *
+     * \since QGIS 3.22
+     */
+    void setFeedback( QgsFeedback *feedback );
+
+    /**
+     * Returns the feedback object that can be queried regularly during rendering to check
+     * if rendering should be canceled, if set. Maybe be NULLPTR.
+     *
+     * \see setFeedback()
+     *
+     * \since QGIS 3.22
+     */
+    QgsFeedback *feedback() const;
+
     /**
      * Returns TRUE if rendering operations should use vector operations instead
      * of any faster raster shortcuts.
@@ -376,12 +393,37 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      */
     double rendererScale() const {return mRendererScale;}
 
+
     /**
-     * Gets access to new labeling engine (may be NULLPTR)
-     * \note not available in Python bindings
+     * Returns the symbology reference scale.
+     *
+     * This represents the desired scale denominator for the rendered map, eg 1000.0 for a 1:1000 map render.
+     * A value of -1 indicates that symbology scaling by reference scale is disabled.
+     *
+     * The symbology reference scale is an optional property which specifies the reference
+     * scale at which symbology in paper units (such a millimeters or points) is fixed
+     * to. For instance, if the scale is 1000 then a 2mm thick line will be rendered at
+     * exactly 2mm thick when a map is rendered at 1:1000, or 1mm thick when rendered at 1:2000, or 4mm thick at 1:500.
+     *
+     * \see setSymbologyReferenceScale()
+     * \see rendererScale()
+     * \since QGIS 3.22
+     */
+    double symbologyReferenceScale() const { return mSymbologyReferenceScale; }
+
+    /**
+     * Gets access to new labeling engine (may be NULLPTR).
+     * \note Not available in Python bindings.
      */
     QgsLabelingEngine *labelingEngine() const { return mLabelingEngine; } SIP_SKIP
 
+    /**
+     * Returns the associated label sink, or NULLPTR if not set.
+     * \note Not available in Python bindings.
+     * \since QGIS 3.24
+     */
+    QgsLabelSink *labelSink() const { return mLabelSink; } SIP_SKIP
+
     /**
      * Returns the color to use when rendering selected features.
      *
@@ -456,14 +498,17 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * Sets whether the rendering operation has been \a stopped and any ongoing
      * rendering should be canceled immediately.
      *
+     * \note Since QGIS 3.22 the feedback() member exists as an alternative means of cancellation support.
+     *
      * \see renderingStopped()
+     * \see feedback()
+     * \see setFeedback()
      */
     void setRenderingStopped( bool stopped ) {mRenderingStopped = stopped;}
 
     /**
      * A general purpose distance and area calculator, capable of performing ellipsoid based calculations.
      * Will be used to convert meter distances to active MapUnit values for QgsUnitTypes::RenderMetersInMapUnits
-     * \since QGIS 3.0
      */
     void setDistanceArea( const QgsDistanceArea &distanceArea ) {mDistanceArea = distanceArea ;}
 
@@ -475,6 +520,14 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      */
     void setScaleFactor( double factor ) {mScaleFactor = factor;}
 
+    /**
+     * Sets the targeted \a dpi for rendering.
+     *
+     * \see dpiTarget()
+     * \since QGIS 3.20
+     */
+    void setDpiTarget( double dpi ) {mDpiTarget = dpi;}
+
     /**
      * Sets the renderer map scale. This should match the desired scale denominator
      * for the rendered map, eg 1000.0 for a 1:1000 map render.
@@ -482,6 +535,23 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      */
     void setRendererScale( double scale ) {mRendererScale = scale;}
 
+    /**
+     * Sets the symbology reference \a scale.
+     *
+     * This should match the desired scale denominator for the rendered map, eg 1000.0 for a 1:1000 map render.
+     * Set to -1 to disable symbology scaling by reference scale.
+     *
+     * The symbology reference scale is an optional property which specifies the reference
+     * scale at which symbology in paper units (such a millimeters or points) is fixed
+     * to. For instance, if \a scale is set to 1000 then a 2mm thick line will be rendered at
+     * exactly 2mm thick when a map is rendered at 1:1000, or 1mm thick when rendered at 1:2000, or 4mm thick at 1:500.
+     *
+     * \see symbologyReferenceScale()
+     * \see rendererScale()
+     * \since QGIS 3.22
+     */
+    void setSymbologyReferenceScale( double scale ) { mSymbologyReferenceScale = scale; }
+
     /**
      * Sets the destination QPainter for the render operation. Ownership of the painter
      * is not transferred and the QPainter destination must stay alive for the duration
@@ -490,6 +560,18 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      */
     void setPainter( QPainter *p ) {mPainter = p;}
 
+    /**
+     * Sets the destination \a painter for temporary in-progress preview renders.
+     * Ownership of \a painter is not transferred and the QPainter destination must
+     * stay alive for the duration of any rendering operations.
+     *
+     * \a painter may be NULLPTR if temporary in-progress preview renders are not required.
+     *
+     * \see previewRenderPainter()
+     * \since QGIS 3.34
+    */
+    void setPreviewRenderPainter( QPainter *painter ) { mPreviewRenderPainter = painter; }
+
     /**
      * Sets a mask QPainter for the render operation. Ownership of the painter
      * is not transferred and the QPainter must stay alive for the duration
@@ -499,17 +581,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      */
     void setMaskPainter( QPainter *p, int id = 0 ) { mMaskPainter[id] = p; }
 
-    /**
-     * When rendering a map layer in a second pass (for selective masking),
-     * some symbol layers may be disabled.
-     *
-     * Sets the list of disabled symbol layers.
-     * \see disabledSymbolLayers()
-     * \see isSymbolLayerEnabled()
-     * \since QGIS 3.12
-     */
-    void setDisabledSymbolLayers( const QSet<const QgsSymbolLayer *> &symbolLayers ) { mDisabledSymbolLayers = symbolLayers; }
-
     /**
      * Sets whether rendering operations should use vector operations instead
      * of any faster raster shortcuts.
@@ -519,11 +590,19 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
     void setForceVectorOutput( bool force );
 
     /**
-     * Assign new labeling engine
-     * \note not available in Python bindings
+     * Assigns the labeling engine
+     * \note Not available in Python bindings.
      */
     void setLabelingEngine( QgsLabelingEngine *engine ) { mLabelingEngine = engine; } SIP_SKIP
 
+    /**
+     * Assigns the label sink which will take over responsibility for handling labels.
+     * \note Ownership is not transferred and the sink must exist for the lifetime of the map rendering job.
+     * \note Not available in Python bindings.
+     * \since QGIS 3.24
+     */
+    void setLabelSink( QgsLabelSink *sink ) { mLabelSink = sink; } SIP_SKIP
+
     /**
      * Sets the \a color to use when rendering selected features.
      *
@@ -560,7 +639,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * The default is to use no simplification.
      *
      * \see setVectorSimplifyMethod()
-     * \since QGIS 2.4
      */
     const QgsVectorSimplifyMethod &vectorSimplifyMethod() const { return mVectorSimplifyMethod; }
 
@@ -575,7 +653,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      *
      * \see vectorSimplifyMethod()
      *
-     * \since QGIS 2.4
      */
     void setVectorSimplifyMethod( const QgsVectorSimplifyMethod &simplifyMethod ) { mVectorSimplifyMethod = simplifyMethod; }
 
@@ -583,7 +660,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * Sets the expression context. This context is used for all expression evaluation
      * associated with this render context.
      * \see expressionContext()
-     * \since QGIS 2.12
      */
     void setExpressionContext( const QgsExpressionContext &context ) { mExpressionContext = context; }
 
@@ -591,7 +667,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * Gets the expression context. This context should be used for all expression evaluation
      * associated with this render context.
      * \see setExpressionContext()
-     * \since QGIS 2.12
      */
     QgsExpressionContext &expressionContext() { return mExpressionContext; }
 
@@ -600,7 +675,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * associated with this render context.
      * \see setExpressionContext()
      * \note not available in Python bindings
-     * \since QGIS 2.12
      */
     const QgsExpressionContext &expressionContext() const { return mExpressionContext; } SIP_SKIP
 
@@ -613,7 +687,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * Set a filter feature provider used for additional filtering of rendered features.
      * \param ffp the filter feature provider
      * \see featureFilterProvider()
-     * \since QGIS 2.14
      */
     void setFeatureFilterProvider( const QgsFeatureFilterProvider *ffp );
 
@@ -621,7 +694,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * Gets the filter feature provider used for additional filtering of rendered features.
      * \returns the filter feature provider
      * \see setFeatureFilterProvider()
-     * \since QGIS 2.14
      */
     const QgsFeatureFilterProvider *featureFilterProvider() const;
 
@@ -658,31 +730,32 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
     /**
      * Converts a size from the specified units to painter units (pixels). The conversion respects the limits
      * specified by the optional scale parameter.
+     *
+     * Since QGIS 3.22 the optional \a property argument can be used to specify the associated property. This
+     * is used in some contexts to refine the converted size. For example, a Qgis::RenderSubcomponentProperty::BlurSize
+     * property will be limited to a suitably fast range when the render context has the Qgis::RenderContextFlag::RenderSymbolPreview set.
+     *
      * \see convertToMapUnits()
-     * \since QGIS 3.0
      */
-    double convertToPainterUnits( double size, QgsUnitTypes::RenderUnit unit, const QgsMapUnitScale &scale = QgsMapUnitScale() ) const;
+    double convertToPainterUnits( double size, Qgis::RenderUnit unit, const QgsMapUnitScale &scale = QgsMapUnitScale(), Qgis::RenderSubcomponentProperty property = Qgis::RenderSubcomponentProperty::Generic ) const;
 
     /**
      * Converts a size from the specified units to map units. The conversion respects the limits
      * specified by the optional scale parameter.
      * \see convertToPainterUnits()
-     * \since QGIS 3.0
      */
-    double convertToMapUnits( double size, QgsUnitTypes::RenderUnit unit, const QgsMapUnitScale &scale = QgsMapUnitScale() ) const;
+    double convertToMapUnits( double size, Qgis::RenderUnit unit, const QgsMapUnitScale &scale = QgsMapUnitScale() ) const;
 
     /**
      * Converts a size from map units to the specified units.
      * \see convertToMapUnits()
-     * \since QGIS 3.0
      */
-    double convertFromMapUnits( double sizeInMapUnits, QgsUnitTypes::RenderUnit outputUnit ) const;
+    double convertFromMapUnits( double sizeInMapUnits, Qgis::RenderUnit outputUnit ) const;
 
     /**
      * Convert meter distances to active MapUnit values for QgsUnitTypes::RenderMetersInMapUnits
      * \note
       * When the sourceCrs() is geographic, the center of the Extent will be used
-     * \since QGIS 3.0
      */
     double convertMetersToMapUnits( double meters ) const;
 
@@ -692,7 +765,7 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * \see setTextRenderFormat()
      * \since QGIS 3.4.3
      */
-    TextRenderFormat textRenderFormat() const
+    Qgis::TextRenderFormat textRenderFormat() const
     {
       return mTextRenderFormat;
     }
@@ -703,7 +776,7 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      * \see textRenderFormat()
      * \since QGIS 3.4.3
      */
-    void setTextRenderFormat( TextRenderFormat format )
+    void setTextRenderFormat( Qgis::TextRenderFormat format )
     {
       mTextRenderFormat = format;
     }
@@ -856,13 +929,191 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
      */
     void setTextureOrigin( const QPointF &origin );
 
+    /**
+     * Add a clip \a path to be applied to the \a symbolLayer before rendering
+     * \since QGIS 3.26, arguments changed and public API since 3.30
+     */
+    void addSymbolLayerClipPath( const QString &symbolLayerId, QPainterPath path );
+
+    /**
+     * Returns clip paths to be applied to the \a symbolLayer before rendering
+     * \since QGIS 3.26, arguments changed and public API since 3.30
+     */
+    QList<QPainterPath> symbolLayerClipPaths( const QString &symbolLayerId ) const;
+
+    /**
+     * Returns the range of z-values which should be rendered.
+     *
+     * \see setZRange()
+     * \since QGIS 3.18
+     */
+    QgsDoubleRange zRange() const;
+
+    /**
+     * Sets the \a range of z-values which should be rendered.
+     *
+     * \see zRange()
+     * \since QGIS 3.18
+     */
+    void setZRange( const QgsDoubleRange &range );
+
+    /**
+     * Returns the size of the resulting rendered image, in pixels.
+     *
+     * \see deviceOutputSize()
+     * \see setOutputSize()
+     *
+     * \since QGIS 3.22
+     */
+    QSize outputSize() const;
+
+    /**
+     * Sets the \a size of the resulting rendered image, in pixels.
+     *
+     * \see outputSize()
+     * \since QGIS 3.22
+     */
+    void setOutputSize( QSize size );
+
+    /**
+     * Returns the device pixel ratio.
+     *
+     * Common values are 1 for normal-dpi displays and 2 for high-dpi "retina" displays.
+     *
+     * \see setDevicePixelRatio()
+     * \since QGIS 3.22
+     */
+    float devicePixelRatio() const;
+
+    /**
+     * Sets the device pixel \a ratio.
+     *
+     * Common values are 1 for normal-dpi displays and 2 for high-dpi "retina" displays.
+     *
+     * \see devicePixelRatio()
+     * \since QGIS 3.22
+     */
+    void setDevicePixelRatio( float ratio );
+
+    /**
+     * Returns the device output size of the render.
+     *
+     * This is equivalent to the output size multiplicated by the device pixel ratio.
+     *
+     * \see outputSize()
+     * \see devicePixelRatio()
+     * \see setOutputSize()
+     *
+     * \since QGIS 3.22
+     */
+    QSize deviceOutputSize() const;
+
+    /**
+     * Sets QImage \a format which should be used for QImages created
+     * during rendering.
+     *
+     * \see imageFormat()
+     * \since QGIS 3.22
+     */
+    void setImageFormat( QImage::Format format ) { mImageFormat = format; }
+
+    /**
+     * Returns the QImage format which should be used for QImages created
+     * during rendering.
+     *
+     * \see setImageFormat
+     * \since QGIS 3.22
+     */
+    QImage::Format imageFormat() const { return mImageFormat; }
+
+    /**
+    * Returns the renderer usage
+    *
+    * \see setRendererUsage()
+    * \since QGIS 3.24
+    */
+    Qgis::RendererUsage rendererUsage() const {return mRendererUsage;}
+
+    /**
+    * Sets the renderer usage
+    *
+    * \note This usage not alter how the map gets rendered but the intention is that data provider
+    * knows the context of rendering and may report that to the backend.
+    *
+    * \see rendererUsage()
+    * \since QGIS 3.24
+    */
+    void setRendererUsage( Qgis::RendererUsage usage ) {mRendererUsage = usage;}
+
+    /**
+     * Returns the frame rate of the map, for maps which are part of an animation.
+     *
+     * Returns -1 if the map is not associated with an animation.
+     *
+     * \see setFrameRate()
+     * \since QGIS 3.26
+     */
+    double frameRate() const;
+
+    /**
+     * Sets the frame \a rate of the map (in frames per second), for maps which are part of an animation.
+     *
+     * Defaults to -1 if the map is not associated with an animation.
+     *
+     * \see frameRate()
+     * \since QGIS 3.26
+     */
+    void setFrameRate( double rate );
+
+    /**
+     * Returns the current frame number of the map (in frames per second), for maps which are part of an animation.
+     *
+     * Returns -1 if the map is not associated with an animation.
+     *
+     * \see setCurrentFrame()
+     * \since QGIS 3.26
+     */
+    long long currentFrame() const;
+
+    /**
+     * Sets the current \a frame of the map, for maps which are part of an animation.
+     *
+     * Defaults to -1 if the map is not associated with an animation.
+     *
+     * \see currentFrame()
+     * \since QGIS 3.26
+     */
+    void setCurrentFrame( long long frame );
+
+
+    /**
+     * Returns the destination elevation map for the render operation.
+     *
+     * \see setElevationMap()
+     * \since QGIS 3.30
+     */
+    QgsElevationMap *elevationMap() const;
+
+    /**
+     * Sets the destination elevation \a map for the render operation. Ownership of the elevation map
+     * is not transferred and the QgsElevationMap destination must stay alive for the duration
+     * of any rendering operations.
+     *
+     * \see elevationMap()
+     * \since QGIS 3.30
+     */
+    void setElevationMap( QgsElevationMap *map );
+
   private:
 
-    Flags mFlags;
+    Qgis::RenderContextFlags mFlags;
 
     //! Painter for rendering operations
     QPainter *mPainter = nullptr;
 
+    //! Painter for in-progress rendering operations
+    QPainter *mPreviewRenderPainter = nullptr;
+
     /**
      * Mask painters for selective masking.
      * Multiple mask painters can be defined for a rendering. The map key is a unique identifier for each mask painter.
@@ -896,7 +1147,6 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
     /**
      * A general purpose distance and area calculator, capable of performing ellipsoid based calculations.
      * Will be used to convert meter distances to active MapUnit values for QgsUnitTypes::RenderMetersInMapUnits
-     * \since QGIS 3.0
      */
     QgsDistanceArea mDistanceArea;
 
@@ -908,15 +1158,26 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
     //! True if the rendering has been canceled
     bool mRenderingStopped = false;
 
+    //! Optional feedback object, as an alternative for mRenderingStopped for cancellation support
+    QgsFeedback *mFeedback = nullptr;
+
     //! Factor to scale line widths and point marker sizes
     double mScaleFactor = 1.0;
 
+    //! Targeted DPI
+    double mDpiTarget = -1.0;
+
     //! Map scale
     double mRendererScale = 1.0;
 
-    //! Newer labeling engine implementation (can be NULLPTR)
+    double mSymbologyReferenceScale = -1;
+
+    //! Labeling engine implementation (can be NULLPTR)
     QgsLabelingEngine *mLabelingEngine = nullptr;
 
+    //! Label sink (can be NULLPTR)
+    QgsLabelSink *mLabelSink = nullptr;
+
     //! Color used for features that are marked as selected
     QColor mSelectionColor;
 
@@ -940,31 +1201,45 @@ class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
 
     QgsPathResolver mPathResolver;
 
-    TextRenderFormat mTextRenderFormat = TextFormatAlwaysOutlines;
+    Qgis::TextRenderFormat mTextRenderFormat = Qgis::TextRenderFormat::AlwaysOutlines;
     QList< QgsRenderedFeatureHandlerInterface * > mRenderedFeatureHandlers;
     bool mHasRenderedFeatureHandlers = false;
     QVariantMap mCustomRenderingFlags;
 
-    QSet<const QgsSymbolLayer *> mDisabledSymbolLayers;
+    QSet<QString> mDisabledSymbolLayers;
 
     QList< QgsMapClippingRegion > mClippingRegions;
     QgsGeometry mFeatureClipGeometry;
 
     QPointF mTextureOrigin;
 
+    QgsDoubleRange mZRange;
+
+    QSize mSize;
+    float mDevicePixelRatio = 1.0;
+    QImage::Format mImageFormat = QImage::Format_ARGB32_Premultiplied;
+
+    Qgis::RendererUsage mRendererUsage = Qgis::RendererUsage::Unknown;
+
+    double mFrameRate = -1;
+    long long mCurrentFrame = -1;
+
+    //! clip paths to be applied to the symbol layer before rendering
+    QMap< QString, QList<QPainterPath> > mSymbolLayerClipPaths;
+
 #ifdef QGISDEBUG
     bool mHasTransformContext = false;
 #endif
-};
 
-Q_DECLARE_OPERATORS_FOR_FLAGS( QgsRenderContext::Flags )
+    QgsElevationMap *mElevationMap = nullptr;
+};
 
 #ifndef SIP_RUN
 
 /**
  * \ingroup core
  *
- * Scoped object for temporary replacement of a QgsRenderContext destination painter.
+ * \brief Scoped object for temporary replacement of a QgsRenderContext destination painter.
  *
  * Temporarily swaps out the destination QPainter object for a QgsRenderContext for the lifetime of the object,
  * before replacing it to the original painter on destruction.
@@ -1020,7 +1295,7 @@ class QgsScopedRenderContextPainterSwap
 /**
  * \ingroup core
  *
- * Scoped object for temporary scaling of a QgsRenderContext for millimeter based rendering.
+ * \brief Scoped object for temporary scaling of a QgsRenderContext for millimeter based rendering.
  *
  * Temporarily scales the destination QPainter for a QgsRenderContext to use millimeter based units for the lifetime of the object,
  * before returning it to pixel based units on destruction.
@@ -1064,7 +1339,7 @@ class QgsScopedRenderContextScaleToMm
 /**
  * \ingroup core
  *
- * Scoped object for temporary scaling of a QgsRenderContext for pixel based rendering.
+ * \brief Scoped object for temporary scaling of a QgsRenderContext for pixel based rendering.
  *
  * Temporarily scales the destination QPainter for a QgsRenderContext to use pixel based units for the lifetime of the object,
  * before returning it to millimeter based units on destruction.
@@ -1108,7 +1383,7 @@ class QgsScopedRenderContextScaleToPixels
 /**
  * \ingroup core
  *
- * Scoped object for saving and restoring a QPainter object's state.
+ * \brief Scoped object for saving and restoring a QPainter object's state.
  *
  * Temporarily saves the QPainter state for the lifetime of the object, before restoring it
  * on destruction.
@@ -1144,6 +1419,59 @@ class QgsScopedQPainterState
     QPainter *mPainter = nullptr;
 };
 
+
+/**
+ * \ingroup core
+ *
+ * \brief Scoped object for temporary override of the symbologyReferenceScale property of a QgsRenderContext.
+ *
+ * Temporarily changes the symbologyReferenceScale, before returning it to the original value on destruction.
+ *
+ * \note Not available in Python bindings
+ * \since QGIS 3.22
+ */
+class QgsScopedRenderContextReferenceScaleOverride
+{
+  public:
+
+    /**
+     * Constructor for QgsScopedRenderContextReferenceScaleOverride.
+     *
+     * Temporarily sets the render \a context symbologyReferenceScale to \a scale for the lifetime of this object.
+     */
+    QgsScopedRenderContextReferenceScaleOverride( QgsRenderContext &context, double scale )
+      : mContext( &context )
+      , mOriginalScale( context.symbologyReferenceScale() )
+    {
+      mContext->setSymbologyReferenceScale( scale );
+    }
+
+    /**
+     * Move constructor.
+     */
+    QgsScopedRenderContextReferenceScaleOverride( QgsScopedRenderContextReferenceScaleOverride &&o ) noexcept
+      : mContext( o.mContext )
+      , mOriginalScale( o.mOriginalScale )
+    {
+      o.mContext = nullptr;
+    }
+
+    /**
+     * Returns the render context back to the original reference scale.
+     */
+    ~QgsScopedRenderContextReferenceScaleOverride()
+    {
+      if ( mContext )
+        mContext->setSymbologyReferenceScale( mOriginalScale );
+    }
+
+  private:
+
+    QgsRenderContext *mContext = nullptr;
+    double mOriginalScale = 0;
+};
+
+
 #endif
 
 #endif
