@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 """
 ***************************************************************************
     BatchOutputSelectionPanel.py
@@ -44,7 +42,7 @@ from processing.gui.AutofillDialog import AutofillDialog
 class BatchOutputSelectionPanel(QWidget):
 
     def __init__(self, output, alg, row, col, panel):
-        super(BatchOutputSelectionPanel, self).__init__(None)
+        super().__init__(None)
         self.alg = alg
         self.row = row
         self.col = col
@@ -57,8 +55,8 @@ class BatchOutputSelectionPanel(QWidget):
         self.text = QLineEdit()
         self.text.setText('')
         self.text.setMinimumWidth(300)
-        self.text.setSizePolicy(QSizePolicy.Expanding,
-                                QSizePolicy.Expanding)
+        self.text.setSizePolicy(QSizePolicy.Policy.Expanding,
+                                QSizePolicy.Policy.Expanding)
         self.horizontalLayout.addWidget(self.text)
         self.pushButton = QPushButton()
         self.pushButton.setText('…')
@@ -87,7 +85,7 @@ class BatchOutputSelectionPanel(QWidget):
                     filename += ext.group(1)
             settings.setValue('/Processing/LastBatchOutputPath', os.path.dirname(filename))
             dlg = AutofillDialog(self.alg)
-            dlg.exec_()
+            dlg.exec()
             if dlg.mode is not None:
                 if dlg.mode == AutofillDialog.DO_NOT_AUTOFILL:
                     self.table.cellWidget(self.row,
@@ -144,7 +142,7 @@ class BatchOutputSelectionPanel(QWidget):
             lastDir = ''
 
         dirName = QFileDialog.getExistingDirectory(self,
-                                                   self.tr('Output Directory'), lastDir, QFileDialog.ShowDirsOnly)
+                                                   self.tr('Output Directory'), lastDir, QFileDialog.Option.ShowDirsOnly)
 
         if dirName:
             self.table.cellWidget(self.row, self.col).setValue(dirName)
